//============================================================================
//
//  Equinox Desktop Environment - Main Panel
//  Copyright (C) 2000 Martin Pekar <cortex@post.sk>
//
//  Requires Fast Light Tool Kit(FLTK)
//  http://www.agilinux.com/fltk
//	
//  Author  : Martin Pekar
//  E-Mail  : cortex@post.sk
//  Home    : http://www.cortex.host.sk

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//============================================================================

#include "Eqpanel.h"
#include "Tasks.h"
#include "WmComm.h"
#include "DeskIcon.h"
#include "PanelButton.h"

//====== V a r i a b l e s   ================================================
Fl_Box 			*clock_box;
Fl_Input 		*filename_field, *name_field, 
			*command_field, *icon_field;
int 			background_color = 220; 
Fl_Color 		c = 220;
Fl_Config 		global_conf;
int 			fd[2];

Fl_Scroll 		*task_scroll;
int 			max_scroll = 0;
int 			scheme_choosed = 0;

PanelButton		*start_button, *kbd_select;
Fl_FileBrowser		*programs_browser;
int			selected, submenu_selected = 0;
Fl_Input 		*filename_field_e, *name_field_e, *command_field_e,
			*icon_field_e;
char 			*scheme_file_path;
char			*yourHomeDir;
Fl_Box 			*color_box, *label_color_box, *selection_color_box, 
			*selection_text_color_box, *highlight_color_box, *off_color_box,
			*text_color_box, *highlight_label_color_box;
Fl_Button 		*color_dialog_button_apply, *color_dialog_button_ok;			
//============================================================================


//====== F u n c t i o n s ===================================================			
void 			run_program(Fl_Widget *w, void* data);

void 			cb_reload(Fl_Widget *, void*);

inline void		scan_items(char *path);
inline void 		scan_icons(char *path);

char*			getHomeDir(); 

static  Fl_Color 	grokColor(const char* configFile, const char *colorString); 

void 			run_app_dialog		(Fl_Widget *, void*);
void 			logout_dialog		(Fl_Widget *, void*);
void 			setup_background_dialog	(Fl_Widget *, void*);
void			menu_editor_dialog	(Fl_Widget *, void*);

void 			cb_change_dir(Fl_Widget *, void*);
void 			cb_directory_up(Fl_Button *, void*);
void 			cb_new_submenu(Fl_Button *, void*);
void 			cb_delete_submenu(Fl_Button *, void*);
void 			cb_new_item(Fl_Button *, void*);
void 			cb_delete_item(Fl_Button *, void*);
void 			cb_edit_item(Fl_Button *, void*);
void 			cb_about_menu_editor(Fl_Widget*, void*);

void 			cbCloseWindow(Fl_Widget*, Fl_Window*); //ok

void 			cb_exit_app(Fl_Widget*, void*);
void 			cb_about(Fl_Widget*, void*);

int  			something_in_dir(char *path);
void			WinManagerRefresh(); //ok

//============================================================================
// ========================================================
void //ok
apply_colors(conf_list keyList, Fl_Box *box, char *schemePathname, char *key)
{
    char configResult[180];
    getconf_list(keyList, key, configResult, sizeof(configResult) );
    Fl_Color schemeColor = grokColor(schemePathname, configResult);
    box->color(schemeColor);
    box->redraw();
}
  
static int //ok
_getprop(Window w, Atom a, Atom type, long len, unsigned char **p)
{
    Atom real_type;
    int format;
    unsigned long n, extra;
    int status;

    status = XGetWindowProperty(fl_display, w, a, 0L, len, False, type, &real_type, &format, &n, &extra, p);
    if (status != Success || *p == 0)
	return -1;
    if (n == 0)
	XFree((char*) *p);
    return n;
}

static bool //ok
getSimpleProperty(Window w, Atom a, long &result)
{
    long *p = 0;

    if (_getprop(w, a, a, 1L, (unsigned char**)&p) <= 0){
	return false;
    }

    result = p[0];
    XFree((char *) p);
    return true;
}

static //ok
Fl_Color grokColor(const char* configFile, const char *colorString) 
{
    char key[80], val[32];
    const char *p = colorString;
    snprintf(key, sizeof(key), "aliases/%s", colorString);
    if (!getconf(configFile, key, val, sizeof(val))) 
	p = val;
    char* q;
    long l = strtoul(p, &q, 0);
    if (!*q) 
	return (Fl_Color)l;
    return fl_rgb(p);
}



void 
save_and_apply(Fl_Widget *, void*)
{
    char *key_global[] = { 
    "color", "label color", "selection color",
    "selection text color", "highlight color", "off color", "text color",
    "highlight label color",
    };
    
    Fl_Box *global_colors_boxes[8] = { 
    color_box, label_color_box, selection_color_box, selection_text_color_box,
    highlight_color_box, off_color_box, text_color_box, highlight_label_color_box
    };

    Fl_Config color_scheme(scheme_file_path);

    for (int boxIndex = 0; boxIndex < 8; boxIndex++) {
        char *locationName = new char[126];
	sprintf(locationName, "widgets/default/%s", key_global[boxIndex]);
	color_scheme.set(locationName, (int) global_colors_boxes[boxIndex]->color());
        delete []locationName;	
    }
    color_scheme.set( "global colors/background", (int) global_colors_boxes[0]->color());

    char *real_scheme = strrchr(scheme_file_path, '/');
    
    int res;  
    Fl_Config flconfig("flconfig");
    if ( (res = flconfig.set("default/scheme", ++real_scheme)) ) 
	puts("Cannot set default scheme.\n");
  

    XEvent ev;
    unsigned int i, nrootwins;
    Window dw1, dw2, *rootwins;
  
    Window Root = RootWindow(fl_display, fl_screen);
    XQueryTree(fl_display, Root, &dw1, &dw2, &rootwins, &nrootwins);

    Atom a = XInternAtom(fl_display, "FLTK_STYLE_WINDOW", False);
    Atom Scheme = XInternAtom(fl_display, "FLTKChangeScheme", False);

    for (i = 0; i < nrootwins; i++) {
	long result = 0;
	getSimpleProperty(rootwins[i], a, result);
        if (result) {
	    ev.xclient.type = ClientMessage;
	    ev.xclient.display = fl_display;
	    ev.xclient.window = rootwins[i];
	    ev.xclient.message_type = Scheme;
	    ev.xclient.format = 32;

        XSendEvent(fl_display, rootwins[i] , False, 0L, &ev);
	}
    }

    XFlush(fl_display);
    XFree((char *) rootwins);
    WinManagerRefresh(); 
}


void 
cb_set_color_to_buttons(Fl_FileBrowser *b, void *scheme) 
{
  conf_list key_list = 0;
  
  char *key_global[] = { 
  "color", "label color", "selection color",
  "selection text color", "highlight color", "off color", "text color",
  "highlight label color",
  };
   
  Fl_Box *global_colors_boxes[8] = { 
  color_box, label_color_box, selection_color_box, selection_text_color_box,
  highlight_color_box, off_color_box, text_color_box, highlight_label_color_box
  };

  char *scheme_home = (char *) scheme;
  const char *item = b->text(b->value());
  
  char *scheme_pathname = new char [ strlen(scheme_home) + 1 + strlen(item) + 3 ];
  sprintf(scheme_pathname, "%s/%s", scheme_home, item );

  scheme_file_path = new char[1024];
  scheme_file_path = scheme_pathname;

  getconf_keys(scheme_pathname, "widgets/default", &key_list);

  for (int ii = 0; ii < 8; ii++)    
  apply_colors(key_list, global_colors_boxes[ii], scheme_pathname, key_global[ii] );

  color_dialog_button_apply->activate();
  color_dialog_button_ok->activate();
}

void 
cb_change_color(Fl_Box *, void *cBox)
{
    Fl_Box *colorBox = (Fl_Box *) cBox;
    Fl_Color oldColor = colorBox->color();
    Fl_Color defColor = oldColor;
  
    fl_color_chooser("Enter color:", defColor);
    if ( defColor != oldColor ) {
	colorBox->color(defColor);
	colorBox->redraw();
    }
}


void 
cb_ok_color_dialog(Fl_Widget *w, void *win)
{
  save_and_apply(0, 0);
  Fl_Window *ww = (Fl_Window *) win;
  ww->hide();
}


void 
addNewScheme(Fl_Widget*, Fl_FileBrowser *fBrowser)
{
    char *schemeName = (char*)fl_input("Enter name of the new scheme");
    if (schemeName) {
        char *schemePath = "/.fltk/schemes";
        char *schemeHome = new char[ strlen( yourHomeDir ) + 1 + strlen(schemePath) + 1 + strlen(schemeName) + 1 + strlen(".scheme") + 1];
        sprintf(schemeHome, "%s%s/%s.scheme", yourHomeDir, schemePath, schemeName);
	Fl_Config schemeConfig(schemeHome);
	schemeConfig.set("description", schemeName);
	schemeConfig.set("version", 1);
	schemeConfig.set("FLTK version", "2.0.0");
	schemeConfig.set("widgets/scrollbar/text box", "down");

	
	char *dirB = (char*)fBrowser->directory();
	fBrowser->directory(dirB);
	delete []schemeHome;
    }
}

       
void 
color_settings_dialog(Fl_Widget*, void*) 
{
    Fl_Window *color_set_dialog = new Fl_Window(555, 335, ("Color propeties") );
    
    Fl_Tabs *tab_matrix = new Fl_Tabs(170, 10, 375, 260);
      
    Fl_Group *global_colors_tab = new Fl_Group(170, 35, 375, 235, ( "Global colors") );
    global_colors_tab->text_box(FL_NO_BOX);
	
    color_box = new Fl_Box(185, 65, 125, 25,  ("Color") );
    color_box->box(FL_UP_BOX);
    color_box->align(FL_ALIGN_TOP | FL_ALIGN_LEFT);
  
    label_color_box = new Fl_Box( 185, 110, 125, 25, _("Label color") );
    label_color_box->box(FL_UP_BOX);
    label_color_box->align(FL_ALIGN_TOP | FL_ALIGN_LEFT);
    
    selection_color_box = new Fl_Box(185, 155, 125, 25, _("Selection color") );
    selection_color_box->box(FL_UP_BOX);
    selection_color_box->align(FL_ALIGN_TOP | FL_ALIGN_LEFT);
    
    selection_text_color_box = new Fl_Box(185, 200, 125, 25, _("Selection text color") );
    selection_text_color_box->box(FL_UP_BOX);
    selection_text_color_box->align(FL_ALIGN_TOP | FL_ALIGN_LEFT);

    off_color_box = new Fl_Box(375, 65, 125, 25, _("Off color"));
    off_color_box->box(FL_UP_BOX);
    off_color_box->align(FL_ALIGN_TOP | FL_ALIGN_LEFT);
   
    highlight_color_box = new Fl_Box(375, 110, 125, 25, _("Highlight color") );
    highlight_color_box->box(FL_UP_BOX);
    highlight_color_box->align(FL_ALIGN_TOP | FL_ALIGN_LEFT);
  
    highlight_label_color_box= new Fl_Box(375, 155, 125, 25, _("Highlight label color") );
    highlight_label_color_box->box(FL_UP_BOX);
    highlight_label_color_box->align(FL_ALIGN_TOP | FL_ALIGN_LEFT);
   
    text_color_box = new Fl_Box(375, 200, 125, 25, _("Text color"));
    text_color_box->box(FL_UP_BOX);
    text_color_box->align(FL_ALIGN_TOP | FL_ALIGN_LEFT);
   
    Fl_Button *button_1 = new Fl_Button(320, 65, 25, 25, "..."); 
    button_1->callback( (Fl_Callback*) cb_change_color, (void *) color_box );
  
    Fl_Button *button_2 = new Fl_Button(320, 110, 25, 25, "..."); 
    button_2->callback( (Fl_Callback*) cb_change_color, (void *) label_color_box );
  
    Fl_Button *button_3 = new Fl_Button(320, 155, 25, 25, "..."); 
    button_3->callback( (Fl_Callback*) cb_change_color, (void *) selection_color_box );  
    
    Fl_Button *button_4 = new Fl_Button(320, 200, 25, 25, "..."); 
     button_4->callback( (Fl_Callback*) cb_change_color, (void *) selection_text_color_box );    
    
    Fl_Button *button_5 = new Fl_Button(510, 65, 25, 25, "..."); 
    button_5->callback( (Fl_Callback*) cb_change_color, (void *) off_color_box );    
    
    Fl_Button *button_6 = new Fl_Button(510, 110, 25, 25, "..."); 
    button_6->callback( (Fl_Callback*) cb_change_color, (void *) highlight_color_box );    
     

    Fl_Button *button_7 = new Fl_Button(510, 155, 25, 25, "..."); 
    button_7->callback( (Fl_Callback*) cb_change_color, (void *) highlight_label_color_box );    
     

    Fl_Button *button_8 = new Fl_Button(510, 200, 25, 25, "..."); 
    button_8->callback( (Fl_Callback*) cb_change_color, (void *) text_color_box );    
    
    global_colors_tab->end();
    tab_matrix->end();

    
    char *spath = "/.fltk/schemes";
    char *schemes_home = new char[ strlen( yourHomeDir ) + 1 + strlen(spath) + 1];
    sprintf(schemes_home, "%s%s", yourHomeDir, spath);

    Fl_FileBrowser *scheme_browser = new Fl_FileBrowser(15, 70, 145, 250, _("Installed schemes:"));
    scheme_browser->filter("*.scheme");  
    scheme_browser->directory(schemes_home);
    scheme_browser->align( FL_ALIGN_TOP | FL_ALIGN_LEFT );
    scheme_browser->callback( (Fl_Callback*) cb_set_color_to_buttons, (void*) schemes_home );
    
    Fl_Button *addSchemeButton = new Fl_Button(30, 20, 105, 25, "Add new scheme");    
    addSchemeButton->callback( (Fl_Callback*) addNewScheme, scheme_browser );          
  
    color_dialog_button_ok = new Fl_Button(205, 295, 85, 21, _("&OK") ); 
    color_dialog_button_ok->callback( (Fl_Callback*) cb_ok_color_dialog, color_set_dialog );      
    color_dialog_button_ok->deactivate();   
    
    color_dialog_button_apply = new Fl_Button(320, 295, 85, 21, _("&Apply") ); 
    color_dialog_button_apply->callback( (Fl_Callback*) save_and_apply );    
    color_dialog_button_apply->deactivate();
     
    Fl_Button *button_cancel = new Fl_Button(430, 295, 85, 21, _("&Cancel") ); 
    button_cancel->callback( (Fl_Callback*) cbCloseWindow, color_set_dialog );
    
    color_set_dialog->end();
    color_set_dialog->show();
}


void 
restart(Fl_Widget *, void*) 
{
  char **argv_exec = new char*[2];
  argv_exec[0] = strdup("eqpanel");
  argv_exec[1] = 0L;
  execv("eqpanel", argv_exec);
  delete [] argv_exec;
  exit(1);
}

void 
cb_browse(Fl_Widget *, Fl_Input *input)
{
     char *f = fl_file_chooser("Choose file ...", "*", input->value() );
     if (f) input->value(f);
}

void 
clock_refresh(void*) 
{
    time_t time_is = time(NULL);
    struct tm *time = localtime(&time_is);
    char *clock_time = new char[10];
    
    char *minutes_time = new char[10];
    sprintf(minutes_time, "%d", time->tm_min);    
    
    char *hours_time = new char[10];
    sprintf(hours_time, "%d", time->tm_hour);    
    
    if (time->tm_min < 10) sprintf(minutes_time, "0%d", time->tm_min);
    if (time->tm_hour < 10) sprintf(hours_time, "0%d", time->tm_hour);
    
    sprintf(clock_time, "%s:%s", hours_time, minutes_time);
    clock_box->label( clock_time );
    
    char *date_info =  new char[128];
    strftime(date_info, 128, "%A, %d. %B %Y", time); 
    clock_box->tooltip(date_info);
    clock_box->parent()->redraw();
    Fl::add_timeout(60, clock_refresh);
}

void 
run_program(Fl_Widget *w, void* data)
{
    char command[1024];
    Fl_Config flconfig( (char* ) data );
    flconfig.get("Desktop Entry/Exec", command, 1024);
    char *exec_staff = new char[ strlen( command ) + 5 ];    
    sprintf(exec_staff, "%s &", command);
    system(exec_staff);
    delete []exec_staff;
}


void 
run_utility(Fl_Widget *w, void* data)
{ 
    system( (char* ) data );
}



int make_new_icon()
{

    char *i = (char*) fl_input("Enter name of the new icon", "The Gimp");
    if (i) {
	    char *iconDir = "/.fltk/desktop";
	    char *iconConfig = new char[ strlen( yourHomeDir ) + 1
		+ strlen(iconDir) + 1 + strlen(i) + 1 + strlen(".desktop") + 1];
	    sprintf(iconConfig, "%s%s/%s.desktop", yourHomeDir, iconDir, i);

	    Fl_Config icfg(iconConfig);
	    icfg.set("Desktop Entry/Icon", "/usr/share/icons/Lightbulb.xpm");
	    icfg.set("Desktop Entry/X", 300);
	    icfg.set("Desktop Entry/Y", 200);
	    icfg.set("Desktop Entry/Name", i);

	    char *u = (char*) fl_input("Program name to execute when icon double clicked", "gimp");
	    icfg.set("Desktop Entry/Exec", u);

	    //printf("%s\n", iconConfig);
	        
            DeskIcon *desktopIcon = new DeskIcon(iconConfig);  
	    desktopIcon->show();
	    sendinfo (fd, "Restart eqwm", 0); 

    }  
    return 0;
  
}



void
create_new_icon_cb(Fl_Widget*, void*)
{
    make_new_icon();
}

void 
create_menu_entries() 
{
    Fl_Item_Group *programs_group, *settings_group;
    Fl_Item *reload_item, *help_item, *run_item, *shutdown_item, 
    *new_icon_item, *menu_edit, *background_edit, *color_edit;

    char *ppath = "/.fltk/programs";

    char *programs_home = new char[ strlen( yourHomeDir ) +1+ strlen(ppath) + 1];
    sprintf(programs_home, "%s%s", yourHomeDir, ppath);
    
    programs_group = new Fl_Item_Group( ("&Programs") );
    programs_group->image( new Fl_Pixmap(folder_big_xpm) );
    scan_items(programs_home);
    
    delete [] programs_home;
    programs_group->end();
    
    settings_group = new Fl_Item_Group( ("&Settings") );
    settings_group->image( new Fl_Pixmap(setup_xpm) );
    
	new_icon_item = new Fl_Item( ("Create new icon") ); 
	new_icon_item->image( new Fl_Pixmap(new_icon_xpm) );
	new_icon_item->callback( (Fl_Callback*) create_new_icon_cb );
    
	menu_edit = new Fl_Item( ("Edit programs menu") ); 
	menu_edit->image( new Fl_Pixmap(edit_programs_xpm) );
	menu_edit->callback( (Fl_Callback*) menu_editor_dialog );

	color_edit = new Fl_Item( ("Color preferences") ); 
	color_edit->image( new Fl_Pixmap(change_colors_xpm) );
	color_edit->callback( (Fl_Callback*) color_settings_dialog );

	
	background_edit = new Fl_Item( ("Desktop background color") ); 
	background_edit->image( new Fl_Pixmap(change_background_xpm) );
	background_edit->callback( (Fl_Callback*) setup_background_dialog);
	
    settings_group->end();

    reload_item = new Fl_Item( ("R&efresh entries") );
    reload_item->image( new Fl_Pixmap(refresh_xpm) );
    reload_item->callback( (Fl_Callback*) cb_reload );

    help_item = new Fl_Item( ("&About Equinox ...") );
    help_item->image( new Fl_Pixmap(help_xpm) );
    help_item->callback( (Fl_Callback*) cb_about );

    run_item = new Fl_Item( ("&Run application ...") );
    run_item->image( new Fl_Pixmap(run_app_xpm) );
    run_item->callback( (Fl_Callback*) run_app_dialog );	    

    new Fl_Divider();

    shutdown_item = new Fl_Item( ("&Logout ...") );
    shutdown_item->image( new Fl_Pixmap(logout_xpm) );
    shutdown_item->callback( (Fl_Callback*) logout_dialog );	    
}

void 
cb_reload(Fl_Widget*, void*)
{
    start_button->clear();     
    start_button->begin();
    create_menu_entries();     
    start_button->end();
}

inline void 
scan_items(char *path)
{
  Fl_Item_Group *new_group;
  Fl_Item       *new_item; 
  dirent        **files;	  
  struct stat   file_info;
  int 	        num_files = 0;

  num_files = filename_list(path, &files);
	
  for (int i = 0; i < num_files; i ++)
   if (strcmp(files[i]->d_name, ".") != 0 && strcmp(files[i]->d_name, "..") != 0)  {
    char *filename = new char[ strlen( path ) + 1 + strlen( files[i]->d_name) + 3]; 	 
    sprintf(filename, "%s/%s", path, files[i]->d_name);
    
    if (filename_isdir(filename)) {
     char *real_name = new char[ strlen(files[i]->d_name) + 3]; 	
     sprintf(real_name, "%s", files[i]->d_name); 
     new_group = new Fl_Item_Group(real_name);
     new_group->image( new Fl_Pixmap(folder_xpm) );
     scan_items(filename);	 
     new_group->end();
    }
  }

  for (int i = 0; i < num_files; i ++) {
   if (strcmp(files[i]->d_name, ".") != 0 && strcmp(files[i]->d_name, "..") != 0) {
    char *filename = new char[ strlen( path ) + 1 + strlen( files[i]->d_name) + 3]; 	 
    sprintf(filename, "%s/%s", path, files[i]->d_name);
	  
    if (!filename_isdir(filename) && filename_match(files[i]->d_name, "*")) {
     char value_name[80], value_icon[80];

     Fl_Config flconfig_name(filename);
     flconfig_name.get("Desktop Entry/Name", value_name, 80);
     char *item_name = new char[ strlen( value_name ) + 1 ];
     strcpy( item_name, value_name );
     new_item = new Fl_Item(item_name); 
     
     Fl_Config flconfig_icon(filename);
     flconfig_icon.get("Desktop Entry/Icon", value_icon, 80);
     char *icon_name = new char[ strlen( value_icon ) + 1 ];
     strcpy( icon_name, value_icon );
	
     new_item->image( new Fl_Pixmap(item_xpm) );
	
     if ( !(stat(icon_name, &file_info) < 0) ) 
      new_item->image( Fl_XPM_Image::get(icon_name) );
     new_item->callback( (Fl_Callback*) run_program, (void *) filename );
    }

   }
   free(files[i]);
  }
  free(files);
}


inline void 
scan_icons(char *path)
{
  DeskIcon 	*desktopIcon;
  dirent        **files;	  
  int 	        num_files = 0;

  num_files = filename_list(path, &files);
	
  for (int i = 0; i < num_files; i ++) {
    if (strcmp(files[i]->d_name, ".") != 0 && strcmp(files[i]->d_name, "..") != 0) {
    char *filename = new char[ strlen( path ) + 1 + strlen( files[i]->d_name) + 3]; 	 
    sprintf(filename, "%s/%s", path, files[i]->d_name);
	  
    if (!filename_isdir(filename) && filename_match(files[i]->d_name, "*.desktop")) {
        desktopIcon = new DeskIcon(filename);  
	desktopIcon->show();
	
    }

   }
   free(files[i]);
  }
  free(files);
}



void 
setup_kbd(Fl_Widget *w, void* d)
{
  char *xmodmap = "xmodmap";
  char *kbd_filename = (char*) d;
  char *to_exec = new char[ strlen( xmodmap ) + 1 + strlen(kbd_filename) + 6]; 	
  sprintf(to_exec, "%s %s &", xmodmap, kbd_filename );
  system(to_exec);  
  delete [] to_exec;
  char *label = strrchr(kbd_filename, '/');
  kbd_select->label(++label);
  kbd_select->tooltip(label);
}

void 
installed_keyboards(char *path) //todo for new
{
 struct dirent *dir_tree;
 DIR           *directory;
 Fl_Item       *new_item; 
 
  if ( !((directory = opendir (path)) == NULL) ) {
    while ( (dir_tree = readdir (directory)) ) {
     if (strcmp (dir_tree->d_name, ".") == 0)  continue;
     if (strcmp (dir_tree->d_name, "..") == 0) continue;
     char *tree_item = dir_tree->d_name;
     char *kbd_name = new char[ strlen(tree_item) + 1]; 	
     strcpy(kbd_name, tree_item);
     new_item = new Fl_Item(kbd_name); 
     char *kbd_filename = new char[ strlen( path ) + 1 + strlen(kbd_name) + 3]; 	
     sprintf(kbd_filename, "%s/%s", path, kbd_name);     
     new_item->callback( (Fl_Callback*) setup_kbd, (void *) kbd_filename );
    }
  closedir(directory);    
  }
}

char* 
getHomeDir()
{
    char *home = 0L; 
    char *var_home = getenv("HOME");
    if (var_home) home = var_home;
    else {
        struct passwd *user_info = getpwuid( getuid() );
        if (user_info) 
	    home = user_info->pw_dir;
    }
    return home;
}


  
void //ok
cbBrowseAll(Fl_Widget *, Fl_Input *input)
{
    char *fileName = fl_file_chooser(_("Select file ..."), "*", input->value());
    if (fileName) { 
	input->value(fileName);
    }
}


Fl_Input *commandInput;

void //ok
runCommand(Fl_Widget*, Fl_Window *windowToClose )
{ 
    char *commandName = (char*)commandInput->value();
    char *execLine = new char[ strlen(commandName) + 3];
    sprintf(execLine, "%s &", commandName);
    system(execLine);
    delete []execLine;
    windowToClose->hide();
}


void 
run_app_dialog(Fl_Widget *, void*) // ok
{
    Fl_Window *runDialogWindow = new Fl_Window(370, 130, _("Run application..."));
    
    commandInput = new Fl_Input(60, 40, 205, 23, _("Enter name of the program\nyou want to run:")) ;
    commandInput->align(FL_ALIGN_TOP_LEFT);
    
    Fl_Button *browseButton = new Fl_Button(275, 40, 80, 23, _("&Find ..."));
    browseButton->callback( (Fl_Callback*) cbBrowseAll, commandInput );
 
    Fl_Box *runDialogIcon = new Fl_Box(10, 15, 35, 35);
    runDialogIcon->image( new Fl_Pixmap(run_app_xpm) );
    
    Fl_Button *okButton = new Fl_Button(60, 90, 80, 23, _("&OK")); 
    okButton->callback( (Fl_Callback*) runCommand, runDialogWindow );
    
    Fl_Button *cancelButton = new Fl_Button(160, 90, 80, 23, _("&Cancel")); 
    cancelButton->callback( (Fl_Callback*) cbCloseWindow, runDialogWindow );

    runDialogWindow->end();
    runDialogWindow->show();
}




 Fl_Radio_Button *log_radio, *restart_radio, *shutdown_radio;

void 
cb_logout(Fl_Widget *, void* )
{ 
  if ( log_radio->value() == 1 ) { 
   sendinfo (fd, "Quit", 0); 
   exit(0);
  }

  if ( restart_radio->value() == 1 ) 
   system( "shutdown -r now" );
  if ( shutdown_radio->value() == 1 ) 
   system( "shutdown -h now" );
}

void 
logout_dialog(Fl_Widget *, void*) 
{
   Fl_Window *logout_win = new Fl_Window( (Fl::w()/2) - (276/2), (Fl::h()/2) - (174/2) ,276, 174, "Logout...");
     { 
     Fl_Box *logout_logo = new Fl_Box(20, 30, 45, 45);
     logout_logo->image( new Fl_Pixmap(logout_xpm) );
     }
     log_radio = new Fl_Radio_Button(75, 20, 115, 22, _(" Logout"));
      log_radio->down_box(FL_ROUND_DOWN_BOX);
      log_radio->value(1);
    
     restart_radio = new Fl_Radio_Button(75, 45, 145, 22, _(" Restart computer"));
      restart_radio->down_box(FL_ROUND_DOWN_BOX);
    
     shutdown_radio = new Fl_Radio_Button(75, 70, 165, 22, _(" Shutdown computer"));
      shutdown_radio->down_box(FL_ROUND_DOWN_BOX);
    
    {
    Fl_Button *okbutton =  new Fl_Button(35, 120, 80, 23, _("&OK"));
    okbutton->callback( (Fl_Callback*) cb_logout );
    }
    {
    Fl_Button *cancelb =  new Fl_Button(160, 120, 80, 23, _("&Cancel"));
    cancelb->callback( (Fl_Callback*) cbCloseWindow, logout_win );
    }
    logout_win->end();
    logout_win->show();
}


void 
choose_background_color(Fl_Widget *b, void *) 
{
    Fl_Color oldColor = c;
 
    fl_color_chooser("Choose desktop background color:", c);
    
    if ( c != oldColor ) {
	Fl_Box *a = (Fl_Box *)b;
	a->color(c);
	a->parent()->redraw();
    }
}


//*************************************************

void 
set_bcolor (int red, int green, int blue)
{
    XColor color;
    XWindowAttributes attributes;
    Window root = RootWindow(fl_display, fl_screen);
    XGetWindowAttributes (fl_display, root, &attributes);

    color.red   = red * 255;
    color.green = blue * 255;
    color.blue  = green * 255;

    color.pixel = BlackPixel(fl_display, root);
  
    if ( !XAllocColor (fl_display, attributes.colormap, &color) ) {
	fprintf (stderr, ("Could not allocate color\n"));
    }

    XSetWindowBackground (fl_display, root, color.pixel);
    XClearWindow (fl_display, root);
}


void 
choose_background_ok(Fl_Widget *, Fl_Window *window) 
{
    uchar r, g, b;
    fl_get_color(c, r, g, b);
    set_bcolor (r, b, g);
    global_conf.set("Default/RootColor", (int) c );
    window->hide();
}

void 
choose_background_apply(Fl_Widget *, void *) 
{
    uchar r, b, g;
    fl_get_color(c, r, g, b);
    set_bcolor (r, b, g);
    global_conf.set("Default/RootColor", (int) c );
}

void 
setup_background_dialog(Fl_Widget *, void *) 
{
    int old_bc_color;
    
    Fl_Window *setup_window = new Fl_Window(305, 310, ("Setup desktop color") );
    
    global_conf.get("Default/RootColor", old_bc_color); c = old_bc_color;
    
    { Fl_Button *o = new Fl_Button(25, 25, 250, 200, _("Desktop"));
      o->box(FL_THIN_UP_BOX);
      o->align(FL_ALIGN_TOP_LEFT);
      o->color( (Fl_Color) old_bc_color );
      o->callback( (Fl_Callback*) choose_background_color );
    }
    { Fl_Box* o = new Fl_Box(60, 225, 180, 20);
      o->box(FL_ENGRAVED_BOX);
      o->color((Fl_Color)32);
    }
    
    Fl_Button *ok_button =  new Fl_Button(25, 275, 80, 23, _("&OK"));
    ok_button->callback( (Fl_Callback*) choose_background_ok , setup_window);

    Fl_Button *apply_button =  new Fl_Button(110, 275, 80, 23, _("&Apply"));
    apply_button->callback( (Fl_Callback*) choose_background_apply );
    
    Fl_Button *cancel_button = new Fl_Button(195, 275, 80, 23, _("&Cancel"));    
    cancel_button->callback( (Fl_Callback*) cbCloseWindow, setup_window  );

    
    setup_window->end();
    setup_window->show();
}


void 
menu_editor_dialog(Fl_Widget*, void*) 
{
    Fl_Window *menu_edit_window = new Fl_Window(345, 270, _("Programs menu properties"));
    Fl_Menu_Bar *menubar = new Fl_Menu_Bar(5, 5, 370, 25); 
        menubar->begin();
	    Fl_Item_Group *file = new Fl_Item_Group(_("&File"));
    		Fl_Item *quit_item = new Fl_Item(_("&Quit"));
//        	quit_item->shortcut(FL_ALT+'Q');
		quit_item->callback(cb_exit_app, menu_edit_window);
	    file->end();
	    Fl_Item_Group *help = new Fl_Item_Group(_("&Help"));
		Fl_Item *about_item = new Fl_Item(_("&About"));
//        	about_item->shortcut(FL_ALT+'A');
     		about_item->callback(cb_about_menu_editor);
	    help->end();
    menubar->end();

    char * pdir = "/.fltk/programs";

    char *prog_path = new char[ strlen(yourHomeDir) + 1 + strlen( pdir ) + 1 ];
    sprintf(prog_path, "%s%s", yourHomeDir, pdir );
    
    programs_browser =  new Fl_FileBrowser(5, 62, 185, 203, _("Programs:"));               
      programs_browser->align(FL_ALIGN_TOP | FL_ALIGN_LEFT);
      programs_browser->type(FL_HOLD_BROWSER);
      programs_browser->directory(prog_path);
      //programs_browser->when(6);
      programs_browser->callback( (Fl_Callback*) cb_change_dir );
    { 
    Fl_Button *go_up = new Fl_Button(140, 35, 49, 14, _("Go up..."));
      go_up->box(FL_THIN_UP_BOX);
      go_up->callback( (Fl_Callback*) cb_directory_up );
      go_up->tooltip(_("Go up one level"));
    }
    { 
    Fl_Button *new_submenu = new Fl_Button(235, 50, 100, 23, _("New submenu"));
      new_submenu->callback( (Fl_Callback*) cb_new_submenu );
    }
    { 
    Fl_Button *del_submenu =  new Fl_Button(235, 80, 100, 23, _("Delete submenu"));
      del_submenu->callback( (Fl_Callback*) cb_delete_submenu );
    }
    { 
    Fl_Button *new_item = new Fl_Button(235, 125, 100, 23, _("New item"));
      new_item->callback( (Fl_Callback*) cb_new_item );
    }
    { 
    Fl_Button *del_item = new Fl_Button(235, 185, 100, 23, _("Delete item"));
    del_item->callback( (Fl_Callback*) cb_delete_item );
    }
    { 
    Fl_Button *edit_item = new Fl_Button(235, 155, 100, 23, _("Edit Item"));
      edit_item->callback( (Fl_Callback*) cb_edit_item );
    }

    menu_edit_window->end();
    menu_edit_window->show();
}


void 
cb_save_item_e(Fl_Widget *a, Fl_Window *w)
{
    const char *item = filename_field_e->value();
    char *dirp = (char*) programs_browser->directory();
    
    char *path_and_item = new char[ strlen(dirp) + 1 + strlen(item) + 1 ];
    sprintf(path_and_item,"%s/%s", dirp, item );
    
    Fl_Config flconfig( path_and_item );
    flconfig.set("Desktop Entry/Name", name_field_e->value() );
    flconfig.set("Desktop Entry/Exec", command_field_e->value());
    flconfig.set("Desktop Entry/Icon", icon_field_e->value());
    delete [] path_and_item;
    programs_browser->directory(dirp);
    programs_browser->relayout();
    cb_reload( 0, 0);
    w->hide();
}



void 
cb_browse_icon(Fl_Widget *w, Fl_Input *input)
{
    char *f = fl_file_chooser("Choose icon file ...", "*.xpm", input->value() );
    if (f)  
	input->value(f);

}


void 
menu_edit_dialog(int edit)
{ 
    Fl_Window *edit_window = new Fl_Window(370, 250, _("Edit Item"));

    filename_field_e = new Fl_Input(5, 25, 195, 23, _("Filename:"));
    filename_field_e->align(FL_ALIGN_TOP | FL_ALIGN_LEFT);
    filename_field_e->value(_("new folder.desktop"));
    
    name_field_e = new Fl_Input(5, 80, 195, 23, _( "Name in the menu:"));
    name_field_e->align(FL_ALIGN_TOP | FL_ALIGN_LEFT);
    name_field_e->value(_("New folder"));
    
    command_field_e = new Fl_Input(5, 125, 195, 23, _("Command to execute:"));
    command_field_e->align(FL_ALIGN_TOP | FL_ALIGN_LEFT);
    command_field_e->value("xterm");
    
    {
    Fl_Button *browse_button_c = new Fl_Button(210, 125, 80, 23, _("F&ind ..."));
    browse_button_c->callback( (Fl_Callback*) cb_browse, command_field_e );
    }
    
    icon_field_e = new Fl_Input(5, 215, 195, 23, _("Icon filename:"));
    icon_field_e->align(FL_ALIGN_TOP | FL_ALIGN_LEFT);
    icon_field_e->value("xterm.xpm");
    
    {
    Fl_Button *save_button = new Fl_Button(285, 25, 80, 23, _("&Save"));
    save_button->callback( (Fl_Callback*) cb_save_item_e, edit_window );
    }
    {
    Fl_Button *cancel_button = new Fl_Button(285, 60, 80, 23, _("&Cancel"));
    cancel_button->callback( (Fl_Callback*) cbCloseWindow, edit_window );
    }
    {
    Fl_Button *browse_button_i = new Fl_Button(210, 215, 80, 23, _("&Find ..."));
    browse_button_i->callback( (Fl_Callback*) cb_browse_icon, icon_field_e );
    }
    if (edit) {
        char* dirp = (char*) programs_browser->directory();
	
	int entries = something_in_dir(dirp);
        if ( entries && selected ) {
		
	int xx = (int)programs_browser->value();
	char *l = (char*)programs_browser->text(xx);
        filename_field_e->value(l);
        char this_value[80];
	const char *tfield = filename_field_e->value();
        char *c_file = new char [ strlen( dirp ) + 1 + strlen( tfield ) + 1];
        sprintf( c_file,"%s/%s", dirp, tfield );
        Fl_Config flconfig( c_file );
        flconfig.get("Desktop Entry/Name", this_value, 80 );
        name_field_e->value(this_value);
        flconfig.get("Desktop Entry/Exec", this_value, 80 );
        command_field_e->value(this_value);
        flconfig.get("Desktop Entry/Icon", this_value, 80 );
        icon_field_e->value(this_value);
	delete [] c_file;
	}
    }
        
    edit_window->end();
    edit_window->show();
}




void 
cb_new_submenu(Fl_Button *, void*)
{
    const char * submenu = 0;
    char *dirp = (char*) programs_browser->directory();
    
    submenu = fl_input( ("Please enter name of the new submenu:") );
    if ( !(submenu == 0) ) {
	char *path_and_submenu = new char[ strlen(dirp) + 1 + strlen(submenu) + 1 ];
	sprintf(path_and_submenu,"%s/%s", dirp, submenu);
	if ( mkdir (path_and_submenu, S_IRUSR | S_IWUSR | S_IXUSR) )
	fl_alert(_("Cannot create submenu!"));
    }
    programs_browser->directory(dirp);  
    programs_browser->relayout();
}



void 
cb_change_dir(Fl_Widget *, void *) 
{
    if (Fl::event_clicks() || Fl::event_key() == FL_Enter) {

	const char *item = programs_browser->text(programs_browser->value());
	char *pathname = new char [ strlen(programs_browser->directory()) + 1 + strlen(item) + 1 ];

	sprintf(pathname, "%s/%s", programs_browser->directory(), item );
	if (filename_isdir(pathname))  {
	    programs_browser->directory(pathname);  
	    programs_browser->relayout();
	    selected = submenu_selected = 0;
    	}   
    }
    else 
        selected = submenu_selected = 1;
}


void 
cb_directory_up(Fl_Button *, void*)
{
    char *adir = "/.fltk/programs";
    char *hd = new char [ strlen(yourHomeDir) + 1 + strlen(adir) + 1 ];
    sprintf(hd, "%s%s", yourHomeDir, adir);

    char *dirp = (char*)programs_browser->directory();
    char *old_path;
    
    if ( !(strcmp(dirp, hd ) == 0) ){ 
	char *backward = strrchr(dirp, '/'); 
	if ( !(backward == NULL) ){
	    int pin = strlen(dirp);
            int bin = strlen(backward);
	    old_path = new char [ strlen(dirp) + 1 ];
            strncpy (old_path, dirp, pin - bin);
            old_path[pin - bin] = '\0';
	    programs_browser->directory(old_path);
	}
    }
    selected = submenu_selected = 0;
 
    delete []hd;
}


int 
something_in_dir(char *path) // todo for new
{
    DIR *directory;
    struct dirent *dir_tree;
    int items = 0;
    directory = opendir (path);
    while ( (dir_tree = readdir (directory)) ) {
	if (strcmp (dir_tree->d_name, ".") == 0) continue;
	if (strcmp (dir_tree->d_name, "..") == 0) continue;
	items++;
    }    
    closedir(directory);
    return items;
}

void 
cb_delete_submenu(Fl_Button *, void*)
{
    struct stat file_info; 
    char *dirp = (char*) programs_browser->directory();
    
    if ( something_in_dir(dirp) && submenu_selected ) {
	const char *item = programs_browser->text(programs_browser->value());
        char *submenu_path = new char[ strlen(dirp) + 1 + strlen(item) + 1 ];
	sprintf(submenu_path, "%s/%s", dirp, item );
    	if ( !(stat(submenu_path, &file_info)) ) {
	    if ( S_ISDIR(file_info.st_mode) ) {
    		if ( something_in_dir(submenu_path) ) 
		fl_alert(_("You should delete all items from submenu,\nbefore you can delete it!"));
    		rmdir (submenu_path);
    		programs_browser->directory(dirp);  
    		programs_browser->relayout();
	    }
	}
	delete [] submenu_path;
    }
}

void 
cb_delete_item(Fl_Button *, void*)
{
    struct stat file_info; 
    char *dirp = (char*) programs_browser->directory();
    
    if ( something_in_dir(dirp) && selected ) {
	const char *item =  programs_browser->text(programs_browser->value());
        char *submenu_path = new char[ strlen(dirp) + 1 + strlen(item) + 1];
	sprintf(submenu_path, "%s/%s", dirp, item );
    	if ( !(stat(submenu_path, &file_info)) ) {
	    if ( !(S_ISDIR(file_info.st_mode)) ) {
		unlink (submenu_path);
    		programs_browser->directory(dirp);  
    		programs_browser->relayout();
	    }
	delete [] submenu_path;
	}
    }
}


void 
cb_new_item(Fl_Button *, void*)
{
    menu_edit_dialog(0); 
}


void 
cb_edit_item(Fl_Button *, void*)
{
    menu_edit_dialog(1); 
}


void 
cb_about_menu_editor(Fl_Widget*, void*) 
{
    fl_message( ("Programs menu editor for Equinox Desktop Environment\n(c) 2000, Martin Pekar\n") );
}

void 
cbCloseWindow(Fl_Widget*, Fl_Window *windowToClose)
{
//    windowToClose->hide();
    delete windowToClose;
}

void 
cb_exit_app(Fl_Widget*w, void*d) 
{
    Fl_Window *t = (Fl_Window*) d;
    t->hide();
}


void 
WinManagerRefresh()  
{
    uchar red, green, blue;
    uchar label_red, label_green, label_blue;
    uchar selection_red, selection_green, selection_blue;
    uchar selection_text_red, selection_text_green, selection_text_blue;
    
    Fl_Color scheme_color, label_color, selection_color, selection_text_color;			
    char config_result[80];
    conf_list key_list = 0;    
  
  
    Fl_Config scheme_current_config("flconfig");


    char *current_scheme =  new char[80];
    scheme_current_config.get("default/scheme", current_scheme, 80 );
    
    conf_clear_cached();
  
    scheme_file_path = new char[1024];
    sprintf(scheme_file_path, "%s/.fltk/schemes/%s", yourHomeDir, current_scheme);
    getconf_keys(scheme_file_path, "widgets/default", &key_list);
  
    getconf_list( key_list, "color", config_result, 80 );
    scheme_color = grokColor(scheme_file_path, config_result);

    getconf_list( key_list, "label color", config_result, 80 );
    label_color = grokColor(scheme_file_path, config_result);
    
    getconf_list( key_list, "selection color", config_result, 80 );
    selection_color = grokColor(scheme_file_path, config_result);

    getconf_list( key_list, "selection text color", config_result, 80 );
    selection_text_color = grokColor(scheme_file_path, config_result);
  

    fl_get_color( scheme_color, red, green, blue );
    fl_get_color( label_color, label_red, label_green, label_blue);
    fl_get_color( selection_color, selection_red, selection_green, selection_blue);    
    fl_get_color( selection_text_color, selection_text_red, selection_text_green, selection_text_blue);        

    char *active_color = new char[80];
    snprintf(active_color, 80, "ActiveColor #%02X%02X%02X #%02X%02X%02X",
	selection_text_red, selection_text_green, selection_text_blue,
	red, green, blue);    
    sendinfo (fd, active_color, 0);
    delete []active_color;

    char *inactive_color = new char[80];
    snprintf(inactive_color, 80,  "InactiveColor #%02X%02X%02X #%02X%02X%02X", 
	label_red, label_green, label_blue, red, green, blue);    
    sendinfo (fd, inactive_color, 0);
    delete []inactive_color;

    char *menu_color = new char[80];
    snprintf(menu_color, 80, "MenuColor #%02X%02X%02X #%02X%02X%02X white #%02X%02X%02X", 
	label_red, label_green, label_blue,
	red, green, blue, selection_red, selection_green, selection_blue);
    sendinfo (fd, menu_color, 0);
    delete []menu_color;
  
    char *title_active_color = new char[80];
    snprintf(title_active_color, 80, "TitleStyle Active Gradient #%02X%02X%02X white",
	selection_red, selection_green, selection_blue);
    sendinfo (fd, title_active_color, 0);
    delete []title_active_color;

    char *title_inactive_color = new char[80];
    snprintf(title_inactive_color, 80, "TitleStyle Inactive Gradient #%02X%02X%02X white",
	selection_red, selection_green, selection_blue);
    sendinfo (fd, title_inactive_color, 0);
    delete []title_inactive_color;
    

}

void
cb_about(Fl_Widget*, void*)
{
    fl_message("Equinox Desktop Environment\nVersion 0.3\n(C) 2000 Martin Pekar\nFast and simple desktop environment");
}



int
main(int argc, char ** argv) 
{
    fl_open_display();
    
    yourHomeDir = getHomeDir();

    max_scroll = Fl::w() - 180;    
    
    if ( argc > 1 ) {
	fd[0] = atoi (argv[1]);
	fd[1] = atoi (argv[2]);
	fcntl (fd[1], F_SETFL, O_NDELAY);
    }

    Fl_Window *top_window = new Fl_Window(0, Fl::h() - 28, Fl::w(), 28);
    top_window->box(FL_UP_BOX);
  
    start_button = new PanelButton(2, 3, 58, 22, "Start"); 
    start_button->begin();
    start_button->set_up_width(204);
    start_button->prelight(1);
    start_button->tooltip( ("Welcome to Equinox Desktop Environment") );
    start_button->image( new Fl_Pixmap(Flower_xpm) );
    start_button->align(FL_ALIGN_INSIDE | FL_ALIGN_LEFT);
    start_button->align(FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
    create_menu_entries(); 
    start_button->end();

//    Fl_Box *devider = new Fl_Box(73, 2, 2, 22);
    //devider->box(FL_EMBOSSED_BOX);
  
    task_scroll =  new Fl_Scroll(77, 2, max_scroll, 24); 
    task_scroll->begin();
    task_scroll->text_box(FL_NO_BOX);
    task_scroll->end();

    Fl_Group *dock = new Fl_Group(Fl::w() - 74, 5, 69, 20);
    dock->box(FL_DOWN_BOX);
    dock->begin();

    kbd_select = new PanelButton(Fl::w() - 71, 7, 20, 16);  

    char *reg_path = "/.fltk/registry/global" ;
    char *global_settings = new char[strlen( yourHomeDir ) + 1 + strlen(reg_path) + 1];
    sprintf(global_settings,"%s/%s", yourHomeDir, reg_path);
    char default_kbd[80] = "/usr/share/keyboards/US" ;
    Fl_Config flconfig(global_settings);
    global_conf = flconfig;
    flconfig.get( "Default/Keyboard", default_kbd, 80 );
    setup_kbd(0, (void*) default_kbd);

    uchar r, b, g;
    flconfig.get( "Default/RootColor", background_color );
    fl_get_color(background_color, r, b, g);
    set_bcolor (r, g, b);

    delete []global_settings;

    kbd_select->begin();
    installed_keyboards( "/usr/share/keyboards" );	    
    kbd_select->box(FL_FLAT_BOX);
    kbd_select->tooltip( _("Click to change keyboard"));  
    kbd_select->set_up_width( 690 );
    kbd_select->prelight( 0 );
    kbd_select->end();

    clock_box = new Fl_Box(Fl::w() - 50, 7, 40, 16);  
    clock_box->box(FL_FLAT_BOX);

    dock->end();

    top_window->clear_border();
    top_window->end();
    top_window->show(1, argv);   

    Fl::add_timeout(0.0, clock_refresh);
    Fl::add_timeout(1, scan_pipe);
  
    if ( argc > 1 ) 
	WinManagerRefresh(); 

    char *desktopIconPath = "/.fltk/desktop";
    char *desktopIconHome = new char[ strlen(yourHomeDir) + 1 + strlen(desktopIconPath) + 1];
    sprintf(desktopIconHome, "%s%s", yourHomeDir, desktopIconPath);
    scan_icons(desktopIconHome);
    delete []desktopIconHome;
  
    return Fl::run();
}



